/*--------------------------------------------------------------*/
/* FILE: comp.pl */

% encode_statement(St, Dict, Code): Code is the translated form of
% statement St with respect to dictionary Dict.

encode_statement(assign(name(X), Expr), Dict, 
	  (ExprCode;
	   instr(store, Addr))
               ):-
	lookup(X, Dict, Addr),
	encode_expr(Expr, Dict, ExprCode).
encode_statement(if(Test, Then, Else), Dict,
	  (TestCode;
	   ThenCode;
	   instr(jump, L2);
	  label(L1);
	   ElseCode;
	  label(L2))
                ):-
	encode_test(Test, Dict, L1, TestCode),
	encode_statement(Then, Dict, ThenCode),
	encode_statement(Else, Dict, ElseCode).
encode_statement(while(Test, Do), Dict,
	  (label(L1);
	    TestCode;
	    DoCode;
	    instr(jump, L1);
	   label(L2))
                ):-
	encode_test(Test, Dict, L2, TestCode),
	encode_statement(Do, Dict, DoCode).
encode_statement(read(name(Name)), Dict, instr(read, Addr)):-
	lookup(Name, Dict, Addr).
encode_statement(write(Expr), Dict,
	  (ExprCode;
	   instr(write, 0))
                ):-
	encode_expr(Expr, Dict, ExprCode).
encode_statement((S1;S2), Dict, (Code1;Code2)):-
	encode_statement(S1, Dict, Code1),
	encode_statement(S2, Dict, Code2).


% encode_expr(Expr, Dict, Code): Code is the translated form of
% expression Expr with respect to dictionary Dict.

encode_expr(Expr, Dict, Code):-
	encode_subexpr(Expr, 0, Dict, Code).


% encode_subexpr(Expr, N, Dict, Code): Code is the translated form of
% expression Expr with respect to dictionary Dict, with auxiliary
% variables below N not used in the code.

encode_subexpr(const(C), _, _, instr(loadc, C)).
encode_subexpr(name(X), _, Dict, instr(load, Addr)):-
	lookup(X, Dict, Addr).
encode_subexpr(expr(Op, Expr1, Expr2), N, Dict,
	  (Expr1Code;
	   Instruction)
		):-
	apply(Op, Expr2, Dict, Instruction), !,
	encode_subexpr(Expr1, N, Dict, Expr1Code).
encode_subexpr(expr(Op, Expr1, Expr2), N, Dict,
	  (Expr2Code;
	   instr(store, Addr);
	   Expr1Code;
	   instr(Opcode, Addr))
		):-
	lookup(N, Dict, Addr),
	encode_subexpr(Expr2, N, Dict, Expr2Code),
	N1 is N+1,
	encode_subexpr(Expr1, N1, Dict, Expr1Code),
	memoryop(Op, Opcode).


% apply(Op, Expr, Dict, Instruction): Expr is a simple expression
% such that the operation 'Acc := Acc Op Expr' can be encoded into a
% single instruction Instruction, with respect to dictionary Dict.

apply(Op, const(C), _, instr(Opcode, C)):-
	literalop(Op, Opcode).
apply(Op, name(Name), Dict, instr(Opcode, Addr)):-
	memoryop(Op, Opcode),
	lookup(Name, Dict, Addr).


% encode_test(Test, Dict, Label, Code): Code is the translated form
% of test Test with respect to dictionary Dict. Code jumps to Label
% if Test is not satisfied.

encode_test(test(Op, Arg1, Arg2), Dict, Label,
	  (ExprCode;
	   instr(JumpIf, Label))
           ):-
	encode_expr(expr(-, Arg1, Arg2), Dict, ExprCode),
	unlessop(Op, JumpIf).


% literalop(Op, Mnem): Mnem is the mnemonic of the instruction 
% Acc := Acc Op <Literal Operand>.

literalop(+, addc).
literalop(-, subc).
literalop(*, mulc).
literalop(/, divc).


% memoryop(Op, Mnem): Mnem is the mnemonic of the instruction 
% Acc := Acc Op <Memory Operand>.

memoryop(+, add).
memoryop(-, sub).
memoryop(*, mul).
memoryop(/, div).


% unlessop(Op, Mnem): Mnem is the mnemonic of the jump instruction 
% jumping if the not(Acc Op 0) relation holds.

unlessop(=, jumpne).
unlessop(<, jumpge).
unlessop(>, jumple).
unlessop(\=, jumpeq).
unlessop(=<, jumpgt).
unlessop(>=, jumplt).


% lookup(Name, Dict, Value): First solution of the goal: the (Name, Value)
% pair is in the dictionary Dict.

lookup(Name, dic(Name, Value, _, _), Value):-!.
lookup(Name, dic(Name1, _, Before, _), Value):-
	Name @< Name1, lookup(Name, Before, Value).
lookup(Name, dic(Name1, _, _, After), Value):-
	Name @> Name1, lookup(Name, After, Value).


% compile(Source, Code): Code is the compiled form of Source.

compile(Source,
	  (Code;
	   instr(halt, 0);
	   block(L))
        ):-
	encode_statement(Source, Dict, Code),
	assemble(Code, 1, N0),
	N1 is N0+1,
	allocate(Dict, N1, N),
	L is N-N1.


% assemble(Code, N0, N): Code can be positioned so that it starts at
% location N0 and the first unused location is N.

assemble((Code1;Code2), N0, N):-
	assemble(Code1, N0, N1),
	assemble(Code2, N1, N).
assemble(instr(_, _), N0, N):-
	N is N0+1.
assemble(label(N), N, N).


% allocate(Dict, N0, N): Locations can be assigned to variables in
% Dict in such a way that the first variable is assigned N0, the next
% N0+1, etc, and the first unused location is N.

allocate(void, N, N):-
	!.
allocate(dic(_Name, N1, Before, After), N0, N):-
	allocate(Before, N0, N1),
	N2 is N1+1,
	allocate(After, N2, N).

