/* Simple lexical analyser */

/* recognises the following tokens:
	TYPE   INITIAL CHAR   SUBSEQUENT CHARS
	name       letter      letter; digit
        integer    digit       digit
        spec       spec        spec
        indiv      indiv       ----
        eof        eof         ----

For the definition of the above char categories see char_type
*/


/* read_tokens(L): L is the list of tokens on the input stream terminated */
/* by end_of_file, or a '.' followed by layout.*/

read_tokens([T|L]):-
	get_code(C), read_token(C, T, C1), read_tokens(C1, T, L).


/* read_tokens(C, T, L): L is the list of tokens on the input stream, */
/* where C is the first character of the input stream and T is the last */
/* token read */

read_tokens(_, end_of_file, []):-
	!.
read_tokens(C, (.), []):-
	char_type(C, layout), !.
read_tokens(C, _, [T|L]) :-
	read_token(C, T, C1), read_tokens(C1, T, L).


/* read_token(C, T, C1): T is the next tokens on the input stream, */
/* where C is the first character of the input stream and C1 is the first  */
/* character following the token T. */

read_token(C, T, C1):-
	char_type(C, Type), read_token(Type, C, T, C1).


/* read_token(Type, C, T, C1): as read_token/3, */
/* but Type is the char_type of C. */

read_token(layout, _, T, C1):-
	!, get_code(C), read_token(C, T, C1).
read_token(eof, _, end_of_file, -1):-
	!.
read_token(Type, C, T, C2):-
	get_code(C1),
	read_token_tail(Type, C1, CL, C2),
	(   Type = digit -> number_codes(T, [C|CL])
	;   atom_codes(T, [C|CL])
	).

/* read_token_tail(Type, C, CL, C1): Given that Type is the char_type of */
/* the first char of the token, C is the next char, CL is the list of */
/* remaining character codes of the next token and C2 is the first char */
/* following the token. */

read_token_tail(Type, C, [C|CL], C2):-
	char_type(C, Type1), continued(Type, Type1), !, get_code(C1),
	  read_token_tail(Type, C1, CL, C2).
read_token_tail(_Type, C, [], C).


/* char_type(C, Type): char code C is of type Type. Possible Types: */
/* digit:		0-9		*/
/* letter:		a-zA-Z_		*/
/* eof:  		code -1		*/
/* layout:		code =< 32	*/
/* indiv:		(),;[]{}	*/
/* spec:		<all others>	*/

char_type(C, Type):-
	char_type0(C, Type0), !, Type = Type0.
char_type(_C, spec).

char_type0(C, digit):-
	C >= 0'0, C =< 0'9.
char_type0(C, letter):-
	C >= 0'a, C =< 0'z ;
	C >= 0'A, C =< 0'Z ;
	C = 0'_.
char_type0(-1, eof).
char_type0(C, layout):-
	C =< 32.
char_type0(C, indiv):-
	indiv_char(C).

indiv_char(0'().
indiv_char(0')).
indiv_char(0',).
indiv_char(0';).
indiv_char(0'[).
indiv_char(0']).
indiv_char(0'{).
indiv_char(0'}).

/* continued(Type, TypeCont): chars of type TypeCont are allowed in  */
/* tokens starting with a char of type Type */

continued(letter, letter).
continued(letter, digit).
continued(digit, digit).
continued(spec, spec).
